(function (global, factory) {
  typeof exports === 'object' && typeof module !== 'undefined' ? module.exports = factory() :
  typeof define === 'function' && define.amd ? define(factory) :
  (global.Arithmetic = factory());
}(this, function () { 'use strict';

  var babelHelpers = {};

  babelHelpers.defineProperty = function (obj, key, value) {
    if (key in obj) {
      Object.defineProperty(obj, key, {
        value: value,
        enumerable: true,
        configurable: true,
        writable: true
      });
    } else {
      obj[key] = value;
    }

    return obj;
  };

  babelHelpers;

  /**
   * Enum of ArithmeticToken Types
   * @typedef TokenType
   * @enum {string}
   */

  var TOKEN_TYPE = {
    ADDITION: "ADDITION",
    SUBTRACTION: "SUBTRACTION",
    MULTIPLICATION: "MULTIPLICATION",
    DIVISION: "DIVISION",
    LEFT_PARENTHESIS: "LEFT_PARENTHESIS",
    RIGHT_PARENTHESIS: "RIGHT_PARENTHESIS",
    VALUE: "VALUE"
  };

  var DIGITS_REGEXP = /[\d\.]/;

  /**
   * @typedef {object} ArithmeticToken
   * @property {TOKEN_TYPE} type
   * @property {number} [value] - associated value when type is TOKEN_TYPE.VALUE
   */

  /**
   * Returns an Array of ArithmeticTokens for given arithmetic expression.
   *
   * @param {string} str - string arithmetic expression
   * @return {ArithmeticToken[]}
   */
  function lexer (str) {
    var tokens = [];
    var i = 0;

    while (i < str.length) {
      var character = str[i];

      if (character.match(DIGITS_REGEXP) !== null) {
        var value = "";
        while (true) {
          value += str[i];
          if (i + 1 === str.length) break;
          if (str[i + 1].match(DIGITS_REGEXP) === null) break;
          // continue consuming characters as long as they
          // are there and are digits.
          i++;
        }

        tokens.push({ type: TOKEN_TYPE.VALUE, value: parseFloat(value) });
      } else {
        switch (character) {
          case " ":
            // no-op
            break;
          case "(":
            tokens.push({ type: TOKEN_TYPE.LEFT_PARENTHESIS });
            break;
          case ")":
            tokens.push({ type: TOKEN_TYPE.RIGHT_PARENTHESIS });
            break;
          case "+":
            tokens.push({ type: TOKEN_TYPE.ADDITION });
            break;
          case "-":
            tokens.push({ type: TOKEN_TYPE.SUBTRACTION });
            break;
          case "*":
            tokens.push({ type: TOKEN_TYPE.MULTIPLICATION });
            break;
          case "/":
            tokens.push({ type: TOKEN_TYPE.DIVISION });
            break;
          case "%":
            {
              // TODO: Consider throwing this into the parser/shunting-yard sort
              if (tokens[tokens.length - 1].type !== TOKEN_TYPE.VALUE) {
                throw new Error("Unsupported use of % (not following a number)");
              }
              var prevNumberToken = tokens.pop();
              tokens.push({ type: TOKEN_TYPE.LEFT_PARENTHESIS }, prevNumberToken, { type: TOKEN_TYPE.DIVISION }, { type: TOKEN_TYPE.VALUE, value: "100" }, { type: TOKEN_TYPE.RIGHT_PARENTHESIS });
              break;
            }
          default:
            {
              throw new Error("Unhandled operator");
            }
        }
      }

      i++;
    }

    return tokens;
  }

  var _PRECEDENCE_MAP;

  var PRECEDENCE_MAP = (_PRECEDENCE_MAP = {}, babelHelpers.defineProperty(_PRECEDENCE_MAP, TOKEN_TYPE.SUBTRACTION, 1), babelHelpers.defineProperty(_PRECEDENCE_MAP, TOKEN_TYPE.ADDITION, 1), babelHelpers.defineProperty(_PRECEDENCE_MAP, TOKEN_TYPE.MULTIPLICATION, 2), babelHelpers.defineProperty(_PRECEDENCE_MAP, TOKEN_TYPE.DIVISION, 2), _PRECEDENCE_MAP);

  /**
   * Sort the input using "Shunting-Yard" algorithm: return input sorted in
   * Reverse Polish Notation order.
   *
   * @param {ArithmeticToken[]} tokens
   * @return {ArithmeticToken[]}
   */
  function shuntingYard (tokens) {
    var output = [];
    var operators = [];

    for (var i = 0; i < tokens.length; i++) {
      var token = tokens[i];

      switch (token.type) {
        case TOKEN_TYPE.VALUE:
          output.push(token);
          break;
        case TOKEN_TYPE.LEFT_PARENTHESIS:
          operators.push(token);
          break;
        case TOKEN_TYPE.RIGHT_PARENTHESIS:
          while (true) {
            if (operators.length === 0) {
              throw new Error("mismatched parenthesis");
            } else if (operators[operators.length - 1].type === TOKEN_TYPE.LEFT_PARENTHESIS) {
              operators.pop();
              break;
            } else {
              output.push(operators.pop());
            }
          }
          break;
        case TOKEN_TYPE.ADDITION:
        case TOKEN_TYPE.SUBTRACTION:
        case TOKEN_TYPE.MULTIPLICATION:
        case TOKEN_TYPE.DIVISION:
          if (operators.length > 0) {
            var tokenPrecedence = PRECEDENCE_MAP[token.type];
            var lastOperatorPrecedence = PRECEDENCE_MAP[operators[operators.length - 1].type];
            // operators precedence is equal to the other one on
            if (lastOperatorPrecedence === tokenPrecedence) output.push(operators.pop());
          }
          operators.push(token);
          break;
        default:
          throw new Error("Unrecognized TOKEN_TYPE: " + token.type);
      }
    }

    while (operators.length > 0) {
      output.push(operators.pop());
    }

    return output;
  }

  function popTuple(stack) {
    return { a: stack.pop(), b: stack.pop() };
  }

  /**
   * Evaluates an array of ArithmeticTokens sorted in Reverse Polish Notation.
   *
   * @param {ArithmeticToken[]} tokens
   * @return {number}
   */
  function evaluator (tokens) {
    var stack = [];

    for (var i = 0; i < tokens.length; i++) {
      var token = tokens[i];

      switch (token.type) {
        case TOKEN_TYPE.VALUE:
          stack.push(token.value);
          break;
        case TOKEN_TYPE.ADDITION:
          {
            var _popTuple = popTuple(stack);

            var a = _popTuple.a;
            var b = _popTuple.b;

            stack.push(a + b);
            break;
          }
        case TOKEN_TYPE.SUBTRACTION:
          {
            var _popTuple2 = popTuple(stack);

            var _a = _popTuple2.a;
            var _b = _popTuple2.b;

            stack.push(_b - _a);
            break;
          }
        case TOKEN_TYPE.MULTIPLICATION:
          {
            var _popTuple3 = popTuple(stack);

            var _a2 = _popTuple3.a;
            var _b2 = _popTuple3.b;

            stack.push(_a2 * _b2);
            break;
          }
        case TOKEN_TYPE.DIVISION:
          {
            var _popTuple4 = popTuple(stack);

            var _a3 = _popTuple4.a;
            var _b3 = _popTuple4.b;

            stack.push(_b3 / _a3);
            break;
          }
        default:
          throw new Error("Unexpected TOKEN_TYPE");
      }
    }

    return stack[0];
  }

  /**
   * Evaluates given arithmetic expression and returns the result.
   *
   * @param {string} arithmeticExpression - e.g. (10% + 10/2 ) * 17
   * @return {number}
   */
  function parse(arithmeticExpression) {
    return evaluator(shuntingYard(lexer(arithmeticExpression)));
  }

  return parse;

}));
//# sourceMappingURL=arithmetic.umd.js.map